/*
Loxone UI Extension v2018-12-14
by Seb Francis (seb303 on https://groups.google.com/forum/#!forum/loxone-english)
*/
$(document).ready(function() {

	if(!("WebSocket" in window)){
		$('#status').html('<span class=error>Browser doesn\'t support WebSockets</span>');
	} else {
		
		// Global variables for interval timers
		var updateChartsInt;
			
		// tabid: boolean
		var rendered_externals = new Object();
		
		// tabid: timeoutId
		var unload_externals = new Object();
		
		connect();
		function connect() {
			
			/*
			Extension of config.lighting
			
			config.lighting[tabid].lights[i].tabid
			config.lighting[tabid].lights[i].lightidx
			config.lighting[tabid].lights[i].controluuid
			config.lighting[tabid].lights[i].subcontroluuid
			config.lighting[tabid].lights[i].controltype	Switch|Pushbutton|Dimmer|Colorpicker|?ColorPickerV2?|LightController|LightControllerV2
			config.lighting[tabid].lights[i].valueuuid	Uuid used when receiving controller values
			
			config.lighting[tabid].lights[i].g		Optional object containing details of group control
			config.lighting[tabid].lights[i].g.controluuid
			config.lighting[tabid].lights[i].g.subcontroluuid
			config.lighting[tabid].lights[i].g.controltype
			config.lighting[tabid].lights[i].g.valueuuid
			
			- for a simple control, valueuuid is found in:
			  controls/.../states/(active|position|color|activeScene|activeMoods)
			- for a subcontrol within a Lighting Controller, valueuuid is found in:
			  controls/.../subControls/.../states/(active|position|color)
			
			config.lighting[tabid].detectors[i].tabid
			config.lighting[tabid].detectors[i].lightidx
			config.lighting[tabid].detectors[i].controluuid
			config.lighting[tabid].detectors[i].controltype	InfoOnlyDigital
			config.lighting[tabid].detectors[i].valueuuid	Uuid used when receiving controller values
			*/
			
			// tabid: boolean
			var rendered = new Object();
			
			// tabid of active tab
			var activeTab = '';
			
			// Map of room name to room uuid
			var roomName2Uuid = new Object();
			
			// Map of category name to cat uuid
			var catName2Uuid = new Object();
			
			// Map of control value uuid to light or detector object
			var uuid2Light = new Object();
			
			// currentValues = {uuid: <value>, ...}
			var currentValues = new Object();
			// onValueChange = {uuid: <event handler>, ...}
			// <event handler> is passed: uuid, newValue
			var onValueChange = new Object();
			
			// Map of value uuid of control back to actual uuid (e.g. for Daytimer controls)
			var valueUuid2Uuid = new Object();
			
			// Loxone structure
			var structure;
			
			// Loxone timezone offset
			var timezoneOffset = 0;
			
			// Browser time of load (ms since epoch)
			var timeOfLoad;
			
			// chartsToKeepUpdated = {<uuid>: <class of html element(s) to insert chart>, ...}
			var chartsToKeepUpdated = new Object();
			
			// Date offset of minicharts <= 0  {<tabid>: <offset, ...}
			var chartDateOffset = new Object();


			
			// delegateObj contains all available delegate methods
			var delegateObj = {
				socketOnDataProgress: function socketOnDataProgress(socket, progress) {
					//log('socketOnDataProgress')
					//log(progress);
				},
				socketOnTokenConfirmed: function socketOnTokenConfirmed(socket, response) {
					//log('socketOnTokenConfirmed');
					//log(response);
				},
				socketOnTokenReceived: function socketOnTokenReceived(socket, result) {
					//log('socketOnTokenReceived');
					//log(result);
				},
				socketOnConnectionClosed: function socketOnConnectionClosed(socket, code) {
					$('#status').html('<span class=error>Miniserver connection closed</span>');
					setTimeout(connect, 1000);
				},
				socketOnEventReceived: function socketOnEventReceived(socket, events, type) {
					switch (type) {
						
					// Event-Table of Value-States
					case 2:
						events.forEach(function(event) {
							currentValues[event.uuid] = event.value;
							if (onValueChange[event.uuid]) {
								onValueChange[event.uuid](event.uuid, event.value);
							}
						});
						break;
						
					// Event-Table of Text-States
					case 3:
						events.forEach(function(event) {
							if ('text' in event) {
								currentValues[event.uuid] = event.text;
								if (onValueChange[event.uuid]) {
									onValueChange[event.uuid](event.uuid, event.text);
								}
							}
						});
						break;
						
					// Event-Table of Daytimer-States
					case 4:
					// Out-Of-Service Indicator
					case 5:
					// Keepalive response
					case 6:
					// Event-Table of Weather-States
					case 7:
						break;
					default:
						log('unhandled socketOnEventReceived, type: '+type);
						log(events);
					}
				}
			};
			
			// deviceUuid is used to identify the device/instance
			var deviceUuid = 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
				var r = Math.random()*16|0, v = c === 'x' ? r : (r&0x3|0x8);
				return v.toString(16);
			});
			// deviceInfo is device specific information like the userAgent of a Browser
			var deviceInfo = window.navigator.userAgent;
			
			// OPTIONAL
			// If no version is set LxCommunicator.WebSocket will fetch the version on its own
			// This version is needed to determine if the Miniserver supports encryption and tokens
			//LxCommunicator.setConfigVersion("9.3.2.20");
			
			// Alias LxCommunicator.WebSocketConfig constructor
			var WebSocketConfig = LxCommunicator.WebSocketConfig;
			
			// Instantiate a config object to pass it to the LxCommunicator.WebSocket later
			var LxConfig = new WebSocketConfig(WebSocketConfig.protocol.WS, deviceUuid, deviceInfo, WebSocketConfig.permission.APP, false);
			
			// OPTIONAL: assign the delegateObj to be able to react on delegate calls
			LxConfig.delegate = delegateObj;
			
			// Instantiate the LxCommunicator.WebSocket, it is our actual WebSocket
			$('#status').html('<span class=error>Connecting to Miniserver</span>');
			var socket = new LxCommunicator.WebSocket(LxConfig);
			
			// Open Websocket connection to miniserver
			var matches = config.credentials.match(/^([^:]*):([^:]*)$/);
			socket.open(config.miniserver, matches[1], matches[2]).then(function() {
				// Send first command
				sendRequest('data/LoxAPP3.json', gotStructure);
			
			}, function(e) {
				$('#status').html('<span class=error>Error: Unable to connect to Miniserver</span>');
				setTimeout(connect, 2000);
			});
		
			// keyboard shortcut handler
			$(document).keypress(handle_keypress);
			$(document).keyup(handle_keyup);
			// registeredKeys = {which: uuid, ...}
			var registeredKeys = [];
			// registeredKeys = {which: keyIdx, ...}
			var registeredKeyIdx = [];


			function gotStructure(msg) {
				$('#status').html('Connected to Miniserver');
				$('#status').append(
					$(' <span id=errorsmsg>Errors occurred</span>').on('mouseover', function() {
						$('#errors').show();
					}).on('mouseout', function() {
						$('#errors').hide();
					})
				);
				
				structure = JSON.parse(msg);
				//log(structure);
				//log(JSON.stringify(structure, null, '\t'));
				//log(currentValues);
				
				for (var uuid in structure.rooms) {
					roomName2Uuid[structure.rooms[uuid].name] = uuid;
				}
				for (var uuid in structure.cats) {
					catName2Uuid[structure.cats[uuid].name] = uuid;
				}
				
				sendRequest('jdev/cfg/timezoneoffset', function(msg) {
					var matches = msg.LL.value.match(/^(\+|-)([0-9]+):/);
					timezoneOffset = parseInt(matches[1]+matches[2]);
					//log('timezoneOffset: '+timezoneOffset);
				});
				sendRequest('jdev/sps/enablebinstatusupdate');
				timeOfLoad = (new Date()).getTime();
				
				renderPage();
				
				if (updateChartsInt) {
					clearInterval(updateChartsInt);
				}
				updateChartsInt = setInterval(updateCharts, 30*60*1000);
			}
			
			function renderPage() {
				var html = '';
				for (var i=0; i<config.tabs.length; i++) {
					html += '<a id="nava'+i+'" href="javascript:void(0);"><span id="nav-'+config.tabs[i].id+'">'+config.tabs[i].name+'</span></a>';
				}
				var currentTab = window.location.hash.substr(3);
				
				$('#nav').html(html);
				var currentIdx = -1;
				var showIdx = 0;
				for (var i=0; i<config.tabs.length; i++) {
					$('#nava'+i).on('click', showTab);
					if (config.tabs[i]['id'] == currentTab) {
						currentIdx = i;
					}
					if (config.tabs[i]['id'] == config.defaultTab) {
						showIdx = i;
					}
				}
				
				if (currentIdx != -1) {
					$('#nava'+currentIdx).trigger('click');
				} else {
					$('#nava'+showIdx).trigger('click');
				}
			}

			function showTab(event) {
				var tab = config.tabs[event.currentTarget.id.substr(4)];
				
				// Start unload timer for externals
				var currentTab = window.location.hash.substr(3);
				if (tab.id != currentTab && config.external[currentTab] && config.external[currentTab].unload) {
					unload_externals[currentTab] = setTimeout(function() { $('#'+currentTab).remove(); delete rendered_externals[currentTab]; }, config.external[currentTab].unload*1000);
				}
				
				$('#nav span').removeClass('active');
				$('#nav-'+tab.id).addClass('active');
				$('.tab').hide();
				$('#'+tab.id).show();
				window.location.hash = '#t-'+tab.id;
				
				activeTab = tab.id;
				switch(tab.type) {
				case 'lighting':
					renderLightingTab(tab.id);
					break;
				case 'general':
					renderGeneralTab(tab.id);
					break;
				case 'external':
					renderExternalTab(tab.id);
					break;
				default:
					errorInfo("Unknown tab type: '"+tab.type+"'");
				}
			}
			
			function renderLightingTab(tabid) {
				if (!rendered[tabid]) {
					var img = config.lighting[tabid].img;
					var html = '<div id='+tabid+' class=tab><div class=lights><div class=overlay></div><img src="'+img+'" width=100%>';
					var img_w = config.lighting[tabid].imgsize[0];
					var img_h = config.lighting[tabid].imgsize[1];
					
					// Lights
					for (var i=0; i<config.lighting[tabid].lights.length; i++) {
						config.lighting[tabid].lights[i].tabid = tabid;
						config.lighting[tabid].lights[i].lightidx = i;
						var light = config.lighting[tabid].lights[i];
						
						// Draw divs
						for (var j=0; j<light.areas.length; j++) {
							var area = light.areas[j];
							switch (area[0]) {
							case 'rect':
								var w = area[3]-area[1];
								var h = area[4]-area[2];
								if (w < 0) {
									var x = area[3];
									w *= -1;
								} else {
									var x = area[1];
								}
								if (h < 0) {
									var y = area[4];
									h *= -1;
								} else {
									var y = area[2];
								}
								x = 100*x/img_w;
								y = 100*y/img_h;
								w = 100*w/img_w;
								h = 100*h/img_h;
								var c = area[5] ? ' '+area[5] : '';
								html += '<div id="'+tabid+'-'+i+'-'+j+'" class="light'+c+'" style="left: '+x+'%; top: '+y+'%; width: '+w+'%; height: '+h+'%;"></div>';
								break;
							case 'circle':
								var r = area[3];
								var x = 100*(area[1]-r)/img_w;
								var y = 100*(area[2]-r)/img_h;
								var w = 200*r/img_w;
								var h = 200*r/img_h;
								var c = area[4] ? ' '+area[4] : '';
								html += '<div id="'+tabid+'-'+i+'-'+j+'" class="light'+c+'" style="left: '+x+'%; top: '+y+'%; width: '+w+'%; height: '+h+'%; border-radius: 50%;"></div>';
								break;
							}
						}
						
						// Find uuid of control & add control data to config.lighting
						var controluuids = lookUpLightingControl(light, tabid+"/"+i);
						if (controluuids.controluuid) {
							config.lighting[tabid].lights[i].controluuid = controluuids.controluuid;
							if (light.subcontrol) {
								config.lighting[tabid].lights[i].subcontroluuid = controluuids.subcontroluuid;
							}
							config.lighting[tabid].lights[i].controltype = controluuids.controltype;
							config.lighting[tabid].lights[i].valueuuid = controluuids.valueuuid;
							
							// Map to light from uuid to handle incoming values
							uuid2Light[controluuids.valueuuid] = config.lighting[tabid].lights[i];
						}
						
						// Do similar lookup for group
						if (light.group) {
							if (!config.lighting[tabid].groups[light.group]) {
								errorInfo("Unknown group '"+light.group+"' in "+tabid+"/"+i+" ("+light.room+")");
							} else {
								var controluuids = lookUpLightingControl(config.lighting[tabid].groups[light.group], tabid+"/group '"+light.group+"'");
								if (controluuids.controluuid) {
									config.lighting[tabid].lights[i].g = {};
									config.lighting[tabid].lights[i].g.controluuid = controluuids.controluuid;
									if (light.subcontrol) {
										config.lighting[tabid].lights[i].g.subcontroluuid = controluuids.subcontroluuid;
									}
									config.lighting[tabid].lights[i].g.controltype = controluuids.controltype;
									config.lighting[tabid].lights[i].g.valueuuid = controluuids.valueuuid;
								}
							}
						}
					}
					var lightidx = i;
					
					// Detectors
					if (!config.lighting[tabid].detectors) {
						config.lighting[tabid].detectors = [];
					}
					for (var i=0; i<config.lighting[tabid].detectors.length; i++) {
						config.lighting[tabid].detectors[i].tabid = tabid;
						config.lighting[tabid].detectors[i].lightidx = lightidx;
						var detector = config.lighting[tabid].detectors[i];
						
						// Draw divs
						for (var j=0; j<detector.areas.length; j++) {
							var area = detector.areas[j];
							switch (area[0]) {
							case 'rect':
								var w = area[3]-area[1];
								var h = area[4]-area[2];
								if (w < 0) {
									var x = area[3];
									w *= -1;
								} else {
									var x = area[1];
								}
								if (h < 0) {
									var y = area[4];
									h *= -1;
								} else {
									var y = area[2];
								}
								x = 100*x/img_w;
								y = 100*y/img_h;
								w = 100*w/img_w;
								h = 100*h/img_h;
								var c = area[5] ? ' '+area[5] : '';
								html += '<div id="'+tabid+'-'+lightidx+'-'+j+'" class="detector'+c+'" style="left: '+x+'%; top: '+y+'%; width: '+w+'%; height: '+h+'%;"></div>';
								break;
							case 'circle':
								var r = area[3];
								var x = 100*(area[1]-r)/img_w;
								var y = 100*(area[2]-r)/img_h;
								var w = 200*r/img_w;
								var h = 200*r/img_h;
								var c = area[4] ? ' '+area[4] : '';
								html += '<div id="'+tabid+'-'+lightidx+'-'+j+'" class="detector'+c+'" style="left: '+x+'%; top: '+y+'%; width: '+w+'%; height: '+h+'%; border-radius: 50%;"></div>';
								break;
							}
						}
						
						// Find uuid of control & add control data to config.lighting
						var controluuids = lookUpLightingControl(detector, tabid+"/"+i);
						if (controluuids.controluuid) {
							config.lighting[tabid].detectors[i].controluuid = controluuids.controluuid;
							config.lighting[tabid].detectors[i].controltype = controluuids.controltype;
							config.lighting[tabid].detectors[i].valueuuid = controluuids.valueuuid;
							
							// Map to detector from uuid to handle incoming values
							uuid2Light[controluuids.valueuuid] = config.lighting[tabid].detectors[i];
						}
						
						lightidx++;
					}
					
					html += '</div></div>';
					
					$('#'+tabid).remove();
					$('#topbar').after(html);
					
					$('#'+tabid)
						.append($('<div id="'+tabid+'-controls" class=controls></div>').on('mouseleave', light_hidecontrols))
						.find('.light').on('mouseenter', light_showcontrols).on('mousedown', light_showcontrols)
						.on('touchstart', light_touchstart);
					rendered[tabid] = true;
					
					$('#'+tabid).show();
					
					// Register handlers for value changes
					for (var i=0; i<config.lighting[tabid].lights.length; i++) {
						var light = config.lighting[tabid].lights[i];
						onValueChange[light.valueuuid] = light_valuechange;
						// If current value exists then trigger handler
						if (currentValues.hasOwnProperty(light.valueuuid)) {
							light_valuechange(light.valueuuid, currentValues[light.valueuuid]);
						}
					}
					for (var i=0; i<config.lighting[tabid].detectors.length; i++) {
						var detector = config.lighting[tabid].detectors[i];
						onValueChange[detector.valueuuid] = light_valuechange;
						// If current value exists then trigger handler
						if (currentValues.hasOwnProperty(detector.valueuuid)) {
							light_valuechange(detector.valueuuid, currentValues[detector.valueuuid]);
						}
					}
				}
			}
			// light = light or group object, with properties: room, cat, control, subcontrol
			// context = context info for error messages
			//
			// Returns object with properties: controluuid, subcontroluuid, controltype, valueuuid (or empty object on error)
			function lookUpLightingControl(light, context) {
				try {
					var roomuuid = roomName2Uuid[light.room];
					if (!roomuuid) {
						throw new Error("Unknown room '"+light.room+"' in "+context);
					}
					if (light.cat) {
						var catuuid = catName2Uuid[light.cat];
						if (!catuuid) {
							throw new Error("Unknown category '"+light.cat+"' in "+context);
						}
					} else {
						var catuuid = '';
					}
					
					var controluuid = false;
					for (var uuid in structure.controls) {
						if (structure.controls[uuid].room == roomuuid && structure.controls[uuid].name == light.control && (!catuuid || structure.controls[uuid].cat == catuuid)) {
							controluuid = uuid;
							break;
						}
					}
					if (!controluuid) {
						throw new Error("Unknown control '"+light.control+"' in "+context+" ("+light.room+")");
					}
					
					
					var control = structure.controls[controluuid];
					
					var subcontroluuid = false;
					if (light.subcontrol) {
						for (var uuid in structure.controls[controluuid].subControls) {
							if (structure.controls[controluuid].subControls[uuid].name == light.subcontrol) {
								subcontroluuid = uuid;
								break;
							}
						}
						if (!subcontroluuid) {
							throw new Error("Unknown subcontrol '"+light.subcontrol+"' in "+context+" ("+light.room+" / "+light.control+")");
						}
						
						control = structure.controls[controluuid].subControls[subcontroluuid];
					}
					
					// Validate control type and determine uuid for controller values
					switch (control.type) {
					case 'Switch':
					case 'Pushbutton':
						var valueuuid = control.states.active;
						break;
					case 'Dimmer':
						var valueuuid = control.states.position;
						break;
					case 'Colorpicker':
					case 'ColorPickerV2':
						var valueuuid = control.states.color;
						break;
					case 'LightController':
						var valueuuid = control.states.activeScene;
						break;
					case 'LightControllerV2':
						var valueuuid = control.states.activeMoods;
						break;
					case 'InfoOnlyAnalog':
					case 'Daytimer':
					case 'ValueSelector':
						var valueuuid = control.states.value;
						break;
					case 'InfoOnlyDigital':
						var valueuuid = control.states.active;
						break;
					default:
						throw new Error("Unhandled control type '"+control.type+"' in "+context+" ("+light.room+")");
					}
					
					return {'controluuid': controluuid, 'subcontroluuid': subcontroluuid, 'controltype': control.type, 'valueuuid': valueuuid};
				}
				catch (e) {
					errorInfo(e);
					return {};
				}
			}
			
			// Event handler when light or detector value update received
			function light_valuechange(uuid, newValue) {
				var light = uuid2Light[uuid];
				var lightdivs = new Array();
				for (var j=0; j<light.areas.length; j++) {
					lightdivs[lightdivs.length] = $('#'+light.tabid+'-'+light.lightidx+'-'+j);
				}
				switch(light.controltype) {
				case 'Switch':
				case 'Pushbutton':
				case 'InfoOnlyDigital':
					if (newValue == 1) {
						for (var i=0; i<lightdivs.length; i++) {
							lightdivs[i].addClass('on');
						}
					} else {
						for (var i=0; i<lightdivs.length; i++) {
							lightdivs[i].removeClass('on');
						}
					}
					break;
				case 'Dimmer':
					if (newValue > 0) {
						for (var i=0; i<lightdivs.length; i++) {
							lightdivs[i].addClass('on').css('background', 'rgba(230, 230, 0, '+(newValue/100)+')');
						}
					} else {
						for (var i=0; i<lightdivs.length; i++) {
							lightdivs[i].removeClass('on').css('background', '#606060');
						}
					}
					// Update current control?
					if ($('.controls').children().attr('data-uuid') == uuid) {
						var yPos = 100-newValue;
						var dimmer = $('.controls').children().children('.fader');
						dimmer.children('.faderForeground').attr('height', yPos);
						dimmer.children('.faderKnob').attr('cy', yPos);
						var swtch = $('.controls').children().children('.switch');
						if (newValue > 0) {
							swtch.addClass('on');
						} else {
							swtch.removeClass('on');
						}
					}
					break;
				case 'Colorpicker':
				case 'ColorPickerV2':
					if (newValue != 'hsv(0,0,0)') {
						var hsl = hsv2hsl(newValue);
						for (var i=0; i<lightdivs.length; i++) {
							lightdivs[i].addClass('on').css('background', hsl);
						}
					} else {
						for (var i=0; i<lightdivs.length; i++) {
							lightdivs[i].removeClass('on').css('background', '#606060');
						}
					}
					break;
				case 'LightController':
					if (newValue != 0) {
						for (var i=0; i<lightdivs.length; i++) {
							lightdivs[i].addClass('on');
						}
					} else {
						for (var i=0; i<lightdivs.length; i++) {
							lightdivs[i].removeClass('on');
						}
					}
					break;
				case 'LightControllerV2':
					newValue = JSON.parse(newValue)[0];
					if (newValue != 0 && newValue != 778) {
						for (var i=0; i<lightdivs.length; i++) {
							lightdivs[i].addClass('on');
						}
					} else {
						for (var i=0; i<lightdivs.length; i++) {
							lightdivs[i].removeClass('on');
						}
					}
					break;
				default:
					errorInfo("Unhandled controltype '"+light.controltype+"' for "+uuid+" / "+newValue);
				}
			}
			
			// Input hsv(0,0->100,0->100)
			// Output: hsl(0->360,0->100%,0->100%)
			function hsv2hsl(hsv) {
				var matches = hsv.match(/hsv\(([0-9]+),([0-9]+),([0-9]+)\)/);
				if (!matches) {
					errorInfo("Invalid hsv colour: '"+hsv+"'");
					return hsl(0,0,0);
				}
				var h = matches[1];
				var s = matches[2];
				var v = matches[3];
				
				var l = (200 - s) * v / 200;
				s = l&&l<100 ? s*v/(l<50 ? l*2 : 200-l*2) : s;
				
				return 'hsl('+h+','+Math.round(s)+'%,'+Math.round(l)+'%)';
			}
			
			function light_touchstart(e) {
				//log('light_touch: '+e.type);
				
				e.preventDefault();
				
				e.type = 'mouseenter';
				e.pageX = e.originalEvent.touches[0].pageX;
				e.pageY = e.originalEvent.touches[0].pageY;
				setTimeout(light_showcontrols, 300, e);
			}
			function light_showcontrols(e) {
				//log('light_showcontrols: '+e.type);
				var matches = e.currentTarget.id.match(/^(.+)-([0-9]+)-([0-9]+)$/);
				var tabid = matches[1];
				var lightidx = matches[2];
				//var areaidx = matches[3];
				
				var light = config.lighting[tabid].lights[lightidx];
				//var html = '<pre>'+JSON.stringify(light, null, '\t')+'</pre>';
				
				// Can't render control if no current value yet
				if (!currentValues.hasOwnProperty(light.valueuuid)) {
					return;
				}
				
				// Has group control?
				if (light.g) {
					var html = $('<div/>').append(light_buildcontrol(light), light_buildcontrol(light.g));
				} else {
					var html = light_buildcontrol(light);
				}

				var controls = $('#'+tabid+'-controls').html(html).show().css({visibility: 'hidden'});
				var elWidth = controls.outerWidth();
				var elHeight = controls.outerHeight();
				var posX = e.pageX-elWidth/2;
				var posY = e.pageY-elHeight/2;
				var wrapper = $(e.currentTarget).parent();
				if (posX+elWidth > wrapper.width()) {
					posX = wrapper.width()-elWidth;
				}
				if (posX < 0) {
					posX = 0;
				}
				if (posY+elHeight > wrapper.height()) {
					posX = wrapper.height()-elHeight;
				}
				if (posY < 0) {
					posY = 0;
				}
				controls.css({left: posX, top: posY, visibility: 'visible'});
			}
			// light = light or group object, with properties: room, cat, control, subcontrol
			// Returns jQuery object of light control, or empty string
			function light_buildcontrol(light) {
				switch (light.controltype) {
				case 'Switch':
					var html = controlElements.switch.click(light, switch_onoff);
					if (currentValues[light.valueuuid] > 0) {
						html.addClass('on');
					} else {
						html.removeClass('on');
					}
					break;
				case 'Pushbutton':
					var html = controlElements.button.click(light, pushbutton_pulse);
					break;					
				case 'Dimmer':
					var dimmer = controlElements.dimmer;
					var yPos = 100-currentValues[light.valueuuid];
					dimmer.children('.faderBackground').on('mousedown touchstart', light, dimmer_jump);
					dimmer.children('.faderForeground').attr('height', yPos).on('mousedown touchstart', light, dimmer_jump);
					dimmer.children('.faderKnob').attr('cy', yPos).on('mousedown touchstart', light, dimmer_grab);
					
					var html = $('<div class=controlgroup data-uuid="'+light.valueuuid+'"/>');
						
					// Don't render onoff if this is a group control
					if (light.lightidx) {
						var onoff = controlElements.switch.click(light, dimmer_onoff);
						if (currentValues[light.valueuuid] > 0) {
							onoff.addClass('on');
						} else {
							onoff.removeClass('on');
						}
					
						html.append(dimmer, onoff);
					} else {
						html.append(dimmer);
					}
					break;
				default:
					var html = '';
				}
				
				if ((light.controltype == 'LightController' || light.controltype == 'LightControllerV2') || (light.subcontroluuid && !light.subcontrolonly)) {
					if (structure.controls[light.controluuid].states.sceneList) {
						// LightController
						var sceneList = (currentValues[structure.controls[light.controluuid].states.sceneList]+',9="All On",0="All Off"').split(',');
						var activeScene = currentValues[structure.controls[light.controluuid].states.activeScene];
					
						var sceneshtml = $('<div class=scenes/>');
						var matches, scene;
						for (var i = 0; i < sceneList.length; i++) {
							if (matches = sceneList[i].match(/([0-9]+)="([^"]*)"/)) {
								scene = $('<div class="scene'+(activeScene==matches[1]?' on':'')+'">'+matches[2]+'</div>').click({light: light, scene: matches[1]}, scene_select);
								sceneshtml.append(scene);
							}
						}
					} else {
						// LightControllerV2
						var sceneList = JSON.parse(currentValues[structure.controls[light.controluuid].states.moodList]);
						var activeScene = JSON.parse(currentValues[structure.controls[light.controluuid].states.activeMoods])[0];
						
						var sceneshtml = $('<div class=scenes/>');
						var matches, scene;
						for (var i = 0; i < sceneList.length; i++) {
							scene = $('<div class="scene'+(activeScene==sceneList[i].id?' on':'')+'">'+sceneList[i].name+'</div>').click({light: light, scene: sceneList[i].id}, scene_selectV2);
							sceneshtml.append(scene);
						}
					}
					
					if (light.subcontroluuid) {
						var html = $('<div class=controlgroup/>').append(html, sceneshtml);
					} else {
						var html = sceneshtml;
					}
				}
				return html;
			}
			function light_hidecontrols(e) {
				//log('light_hidecontrols: '+e.type);
				$('.controls').hide().html('');
			}
			
			function switch_onoff(e) {
				//log('switch_onoff: '+e.type);
				var light = e.data;
				if (currentValues[light.valueuuid] == 0) {
					var cmd = 'on';
					$(e.currentTarget).addClass('on');
				} else {
					var cmd = 'off';
					$(e.currentTarget).removeClass('on');
				}
				sendRequest('jdev/sps/io/'+(light.subcontroluuid?light.subcontroluuid:light.controluuid)+'/'+cmd);
			}
			
			function pushbutton_pulse(e) {
				//log('pushbutton_pulse: '+e.type);
				var light = e.data;
				$(this).addClass('on');
				var el = this;
				setTimeout(function() {
					$(el).removeClass('on');
				}, 1000);
				sendRequest('jdev/sps/io/'+(light.subcontroluuid?light.subcontroluuid:light.controluuid)+'/pulse');			
			}
			
			function dimmer_onoff(e) {
				//log('dimmer_onoff: '+e.type);
				var light = e.data;
				if (currentValues[light.valueuuid] == 0) {
					var cmd = 'on';
					$(e.currentTarget).addClass('on');
				} else {
					var cmd = 'off';
					$(e.currentTarget).removeClass('on');
				}
				sendRequest('jdev/sps/io/'+(light.subcontroluuid?light.subcontroluuid:light.controluuid)+'/'+cmd);
			}
			
			var dimmerOffset;
			function dimmer_jump(e) {
				//log('dimmer_jump: '+e.type);
				if (e.type == 'touchstart') {
					e.pageY = e.originalEvent.touches[0].pageY;
				}
				var newValue = 100-e.pageY+controlElements['dimmer'].children('.faderBackground').position().top;
				if (newValue < 0) {
					newValue = 0;
				} else if (newValue > 100) {
					newValue = 100;
				}
				var light = e.data;
				sendRequest('jdev/sps/io/'+(light.subcontroluuid?light.subcontroluuid:light.controluuid)+'/'+newValue);
				var yPos = 100-newValue;
				controlElements['dimmer'].children('.faderKnob').attr('cy', yPos).trigger('mousedown', e);
				controlElements['dimmer'].children('.faderForeground').attr('height', yPos);
			}
			function dimmer_grab(e, e2) {
				//log('dimmer_grab: '+e.type);
				if (typeof e2 == 'undefined') {
					e2 = e;
				}
				if (e.type == 'touchstart') {
					e.pageY = e.originalEvent.touches[0].pageY;
				}
				dimmerOffset = 100 - parseInt($(e.currentTarget).attr('cy')) + e2.pageY;
				controlElements['dimmer'].on('mousemove touchmove', e.data, dimmer_drag);
				$(document).on('mouseup touchend', dimmer_release);
			}
			function dimmer_drag(e) {
				//log('dimmer_drag: '+e.type);
				if (e.type == 'touchmove') {
					e.preventDefault();
					e.pageY = e.originalEvent.touches[0].pageY;
				}
				var newValue = dimmerOffset-e.pageY;
				if (newValue < 0) {
					newValue = 0;
				} else if (newValue > 100) {
					newValue = 100;
				}
				var light = e.data;
				sendRequest('jdev/sps/io/'+(light.subcontroluuid?light.subcontroluuid:light.controluuid)+'/'+newValue);
				var yPos = 100-newValue;
				controlElements['dimmer'].children('.faderKnob').attr('cy', yPos);
				controlElements['dimmer'].children('.faderForeground').attr('height', yPos);
			}
			function dimmer_release(e) {
				//log('dimmer_release: '+e.type);
				controlElements['dimmer'].off('mousemove');
				$(document).off('mouseup');
			}
			
			function scene_select(e) {
				//log('scene_select: '+e.type);
				var light = e.data.light;
				var scene = e.data.scene;
				$(e.currentTarget).siblings().removeClass('on');
				$(e.currentTarget).addClass('on');
				if (scene == 9) {
					scene = 'on';
				}
				sendRequest('jdev/sps/io/'+light.controluuid+'/'+scene);
			}
			function scene_selectV2(e) {
				//log('scene_selectV2: '+e.type);
				var light = e.data.light;
				var scene = e.data.scene;
				$(e.currentTarget).siblings().removeClass('on');
				$(e.currentTarget).addClass('on');
				sendRequest('jdev/sps/io/'+light.controluuid+'/changeTo/'+scene);
			}
			
			
			function renderGeneralTab(tabid) {
				if (!rendered[tabid]) {
					var html = '<div id='+tabid+' class=tab>'+config.general[tabid].html+'</div>';
					
					$('#'+tabid).remove();
					$('#topbar').after(html);
					
					var now = new Date();
					var day = ('0'+now.getDate()).substr(-2);
					var month = ('0'+(now.getMonth()+1)).substr(-2);
					var year = now.getFullYear().toString();
					var chartDate = year+month+day;
					chartDateOffset[tabid] = 0;
					
					$('#'+tabid).find("[data-looproom]").each(function() {
						var rooms = $(this).attr('data-looproom').split(',');
						$(this).removeAttr('data-looproom');
						//$(this).find("[data-room]").attr('data-forceroom','');
						for (var i=0; i<rooms.length; i++) {
							$(this).find("[data-thisroom]").html(rooms[i]);
							$(this).find("[data-control]").attr('data-room', rooms[i]);
							//$(this).find("[data-control]").not("[data-forceroom]").attr('data-room', rooms[i]);
							$(this).clone().insertBefore(this);
						}
						$(this).remove();
					});
					$('#'+tabid).find("[data-control]").each(function() {
						switch ($(this).attr('data-display')) {
						case undefined:
						case '':
						case 'value':
						case 'bar':
							var uuid = findControlUuid($(this).attr('data-control'), $(this).attr('data-room'), $(this).attr('data-cat'), $(this).attr('data-suppresserrors'));
							if (uuid == '') {
								return;
							}
							var control = structure.controls[uuid];
							$(this).addClass(uuid);
							
							switch (control.type) {
							case 'Dimmer':
								var valueUuid = control.states.position;
								break;
							case 'Colorpicker':
							case 'ColorPickerV2':
								var valueUuid = control.states.color;
								break;
							case 'LightController':
								var valueUuid = control.states.activeScene;
								break;
							case 'LightControllerV2':
								var valueUuid = control.states.activeMoods;
								break;
							case 'InfoOnlyAnalog':
							case 'Daytimer':
							case 'ValueSelector':
								var valueUuid = control.states.value;
								break;
							case 'InfoOnlyDigital':
							case 'Switch':
							case 'Pushbutton':
								var valueUuid = control.states.active;
								break;
							case 'Radio':
								var valueUuid = control.states.activeOutput;
								break;
							}
							switch (control.type) {
							case 'Dimmer':
							case 'Colorpicker':
							case 'ColorPickerV2':
							case 'LightController':
							case 'LightControllerV2':
							case 'InfoOnlyAnalog':
							case 'Daytimer':
							case 'ValueSelector':
							case 'InfoOnlyDigital':
							case 'Switch':
							case 'Pushbutton':
							case 'Radio':
								valueUuid2Uuid[valueUuid] = uuid;
								uuid = valueUuid;
								break;
							}
							
							onValueChange[uuid] = general_updateValue;
							// If current value exists then trigger handler
							if (currentValues.hasOwnProperty(uuid)) {
								general_updateValue(uuid, currentValues[uuid]);
							}
							
							// Input elements act as controls
							if ($(this).prop('tagName') == 'INPUT' || $(this).prop('tagName') == 'SELECT') {
								var keys = $(this).attr('data-keys');
								if (keys != undefined && keys != '') {
									switch(keys) {
									case 'number':
										//registeredKeys[46] = uuid;
										//registeredKeys[48] = uuid;
										registeredKeys[49] = uuid;
										registeredKeys[50] = uuid;
										registeredKeys[51] = uuid;
										registeredKeys[52] = uuid;
										registeredKeys[53] = uuid;
										registeredKeys[54] = uuid;
										registeredKeys[55] = uuid;
										registeredKeys[56] = uuid;
										registeredKeys[57] = uuid;
										break;
									case 'enter':
										registeredKeys[13] = uuid;
										break;
									default:
										var keysUpper = keys.toUpperCase();
										for (var i = 0; i < keys.length; i++) {
											var charCode = keys.charCodeAt(i);
											registeredKeys[charCode] = uuid;
											registeredKeyIdx[charCode] = i;
											
											var charCode = keysUpper.charCodeAt(i);
											registeredKeys[charCode] = uuid;
											registeredKeyIdx[charCode] = i;
										}
									}
								}
								
								$(this).attr('id', uuid);
								switch($(this).prop('tagName')) {
								case 'INPUT':
									switch ($(this).attr('type')) {
									case 'text':
										$(this).change({uuid:uuid}, general_controlChanged).focus(function() {
											$(this).select();
										}).keypress(function(e) {
											e.stopPropagation();
										});
										break;
									case 'radio':
										if (control.details.outputs) {
											var container = $('<div id="'+uuid+'-Container"></div>');
											for (var idx = 1; idx <= 16; idx++) {
												var txt = control.details.outputs[idx];
												if (txt == '' || txt == undefined) {
													break;
												}
												txt = '<span class=radiolabeltext>'+txt+'</span>';
												if (keys != undefined && keys != '') {
													txt += ' <span class=radioshortcut>'+keys[idx-1].toUpperCase()+'</span>';
												}
												var opt = $('<div><label for="'+uuid+idx+'">'+txt+'</label></div>');
												$(this).clone()
													.attr({id:uuid+idx, name:uuid, value:idx})
													.change({uuid:uuid}, general_controlChanged)
													.prependTo(opt);
												opt.appendTo(container);
											}
										}
										$(this).replaceWith(container);
										break;
									case 'range':
										if (control.states && control.states.min) {			
											$(this).attr('min', currentValues[control.states.min])
												.attr('max', currentValues[control.states.max])
												.attr('step', currentValues[control.states.step]);
											var rangeEl = this;
											onValueChange[control.states.min] = function(uuid, newValue) {
												$(rangeEl).attr('min', newValue);
											};
											onValueChange[control.states.max] = function(uuid, newValue) {
												$(rangeEl).attr('max', newValue);
											};
											onValueChange[control.states.step] = function(uuid, newValue) {
												$(rangeEl).attr('step', newValue);
											};
										} else if (control.details) {
											$(this).attr('min', control.details.min)
												.attr('max', control.details.max)
												.attr('step', control.details.step);
										} else {
											$(this).attr('min', 0)
												.attr('max', 100)
												.attr('step', 1);
										}
										$(this).on('input', function outputUpdate(e) {
												var matches;
												if (!control.details) {
													var value = e.currentTarget.value+'%';
												} else if (matches = control.details.format.match(/%.([0-9]+)f(.*)/)) {
													matches[2] = matches[2].replace('%%', '%');
													var value = parseFloat(e.currentTarget.value).toFixed(matches[1])+matches[2];
												} else {
													var value = e.currentTarget.value;
												}
												var uuid = e.currentTarget.id;
												if (valueUuid2Uuid[uuid]) {
													uuid = valueUuid2Uuid[uuid];
												}
												$('#'+uuid+'-Output').val(value);
												
												var lastvalue = $(e.currentTarget).attr('data-lastvalue');
												if (lastvalue !== undefined && value != lastvalue) {
													e.data = {uuid:uuid};
													general_controlChanged(e);
												}
												$(e.currentTarget).attr('data-lastvalue', value);
											})
											.trigger('input');
										break;
									case 'checkbox':
										$(this).change({uuid:uuid}, general_controlChanged);
										break;
									case 'button':
										$(this).mousedown({uuid:uuid}, general_buttonDown)
											.mouseup({uuid:uuid}, general_buttonUp);
										break;
									}
									break;
									
								case 'SELECT':
									// Populate drop down from scene list of Lighting Controller
									if (control.type == 'LightController') {
										// LightController
										onValueChange[control.states.sceneList] = function(uuid_, newValue) {
											general_populateSceneListSelects(valueUuid2Uuid[uuid], newValue);
										};
										general_populateSceneListSelects(valueUuid2Uuid[uuid], currentValues[control.states.sceneList]);
									} else if (control.type == 'LightControllerV2') {
										// LightControllerV2
										onValueChange[control.states.moodList] = function(uuid_, newValue) {
											general_populateSceneListSelects(valueUuid2Uuid[uuid], newValue);
										};
										general_populateSceneListSelects(valueUuid2Uuid[uuid], currentValues[control.states.sceneList]);
									}
									break;
								}
							}
							break;
						case 'chart':
							var uuid = findControlUuid($(this).attr('data-control'), $(this).attr('data-room'), $(this).attr('data-cat'), $(this).attr('data-suppresserrors'));
							if (uuid == '') {
								return;
							}
							
							$(this).addClass(uuid+'-chart').click({uuid:uuid}, showLargeChart).css('cursor','pointer');
							var cmd = 'binstatisticdata/'+uuid+'/'+chartDate;
							var destination = uuid+'-chart';
							//loadBin(cmd, 'minichart', destination); <- Call updateCharts at end instead
							chartsToKeepUpdated[uuid] = destination;
							break;
						case 'output':
							// Used in connection with input range element
							var uuid = findControlUuid($(this).attr('data-control'), $(this).attr('data-room'), $(this).attr('data-cat'), $(this).attr('data-suppresserrors'));
							if (uuid == '') {
								return;
							}
							
							$(this).attr('id', uuid+'-Output').attr('for', uuid);
							break;
						default:
							errorInfo("Unknown data-display '"+$(this).attr('data-display')+"' for control: '"+$(this).attr('data-control')+"'");
						}
					});
					$('#'+tabid).find("[data-chartrange]").addClass('chart-time-range');
					$('#'+tabid).find("[data-chartcontrols]").addClass('chart-controls');
					rendered[tabid] = true;
					setTimeout(updateCharts, 1000);
					
					$('#'+tabid).show();
				}
			}
			
			function general_populateSceneListSelects(uuid, newValue) {
				$('.'+uuid).each(function() {
					if ($(this).prop('tagName') == 'SELECT') {
						$(this).empty();
						var control = structure.controls[uuid];
						if (control.type == 'LightController') {
							// LightController
							var sceneList = ('0="All Off",'+newValue+',9="All On"').split(',');
							var activeScene = currentValues[control.states.activeScene];
						
							var matches, scene;
							for (var i = 0; i < sceneList.length; i++) {
								if (matches = sceneList[i].match(/([0-9]+)="([^"]*)"/)) {
									scene = $('<option'+(activeScene==matches[1]?' selected':'')+' value='+matches[1]+'>'+matches[2]+'</option>');
									$(this).append(scene);
								}
							}
						} else if (control.type == 'LightControllerV2') {
							// LightControllerV2
							var sceneList = JSON.parse(newValue);
							var activeScene = JSON.parse(currentValues[control.states.activeMoods])[0];
							
							var sceneshtml = $('<div class=scenes/>');
							var matches, scene;
							for (var i = 0; i < sceneList.length; i++) {
								scene = $('<option'+(activeScene==sceneList[i].id?' selected':'')+' value='+sceneList[i].id+'>'+sceneList[i].name+'</option>');
								$(this).append(scene);
							}
						}
						$(this).off('change').change({uuid:uuid}, general_controlChanged);
					}
				});
			}
			function general_updateValue(uuid, newValue) {
				if (valueUuid2Uuid[uuid]) {
					uuid = valueUuid2Uuid[uuid];
					bRedirect = false;
				} else {
					bRedirect = true;
				}
				$('.'+uuid).each(function() {
					if (structure.controls[uuid].type == 'Pushbutton') {
						var html = controlElements.button.clone().attr('id',uuid).click({uuid:uuid}, general_controlClicked);
						
					} else if (structure.controls[uuid].type == 'Switch') {
						var active = structure.controls[uuid].states.active;
						valueUuid2Uuid[active] = uuid;
						if ($(this).prop('tagName') != 'INPUT') {
							onValueChange[active] = general_updateValue2;
							var html = $('<div class="switch'+(currentValues[active]?' on':'')+'" id="'+uuid+'"><div></div></div>').click({uuid:uuid}, general_controlClicked);
						}
						
					} else if (structure.controls[uuid].type == 'Alarm' && bRedirect) {
						var armed = structure.controls[uuid].states.armed;
						valueUuid2Uuid[armed] = uuid;
						onValueChange[armed] = general_updateValue;
						// If current value exists then trigger handler
						if (currentValues.hasOwnProperty(armed)) {
							general_updateValue(armed, currentValues[armed]);
						}
						return;
						
					} else if (typeof $(this).attr('data-function') != 'undefined') {
						var html = config[$(this).attr('data-function')](newValue);
						
					} else if (!structure.controls[uuid].details) {
						if ($(this).attr('data-display') == 'bar') {
							var barscale = $(this).attr('data-barscale');
							if (typeof barscale == 'undefined') {
								barscale = 1;
							}
							var html = '<div class=bar style="width:'+(parseFloat(newValue)*barscale)+'px;"></div><div class=barvalue>'+newValue+'</div>';
						} else {
							var html = newValue;
						}
						
					} else if (structure.controls[uuid].details.format) {
						var matches;
						if (matches = structure.controls[uuid].details.format.match(/%.([0-9]+)f(.*)/)) {
							matches[2] = matches[2].replace('%%', '%');
							var value = parseFloat(newValue).toFixed(matches[1])+matches[2];
						} else {
							var value = newValue;
						}
						if (typeof $(this).attr('data-addplus') != 'undefined' && value.substring(0,1) != '-') {
							value = '+'+value;
						}
						if ($(this).attr('data-display') == 'bar') {
							var barscale = $(this).attr('data-barscale');
							if (typeof barscale == 'undefined') {
								barscale = 1;
							}
							var html = '<div class=bar style="width:'+(parseFloat(value)*barscale)+'px;"></div><div class=barvalue>'+value+'</div>';
						} else {
							var html = value;
						}
						
					} else if (structure.controls[uuid].details.text) {
						if (newValue == 0) {
							var html = '<span style="color:'+structure.controls[uuid].details.color.off+'">'+structure.controls[uuid].details.text.off+'</span>';
						} else {
							var html = '<span style="color:'+structure.controls[uuid].details.color.on+'">'+structure.controls[uuid].details.text.on+'</span>';
						}
						
					} else if (structure.controls[uuid].details.image) {
						if (newValue == 0) {
							loadBin(structure.controls[uuid].details.image.off+'.svg', 'svg', uuid);
						} else {
							loadBin(structure.controls[uuid].details.image.on+'.svg', 'svg', uuid);
						}
						return;
						
					} else {
						var html = newValue;
					}
					// Input elements
					switch($(this).prop('tagName')) {
					case 'INPUT':
						switch ($(this).attr('type')) {
						case 'text':
							if (structure.controls[uuid].details.format) {
								matches = html.match(/^([0-9.]+)(.*)$/);
								$(this).val(matches[1]);
								$(this).next('.afterbit').remove();
								$(this).after('<span class=afterbit>'+matches[2]);
							} else {
								$(this).val(html);
							}
							break;
						case 'radio':
							if ($(this).attr('value') == newValue) {
								$(this).prop('checked', true);
							}
							break;
						case 'range':
							$(this).prop('value', newValue).trigger('input');
							break;
						case 'checkbox':
							$(this).prop('checked', newValue);
							break;
						case 'button':
							if (newValue) {
								$(this).addClass('active');
							} else {
								$(this).removeClass('active');
							}
							break;
						}
						break;
						
					case 'SELECT':
						var control = structure.controls[uuid];
						if (control.type == 'LightController' || control.type == 'LightControllerV2') {
							this.value = newValue;
						}
						break;
						
					default:
						$(this).html(html);
					}
				});
			}
			function general_updateValue2(uuid, newValue) {
				if (valueUuid2Uuid[uuid]) {
					uuid = valueUuid2Uuid[uuid];
				}
				switch (structure.controls[uuid].type) {
				case 'Switch':
					if (newValue) {
						$('#'+uuid).addClass('on');
					} else {
						$('#'+uuid).removeClass('on');
					}
					break;
				}
			}
			
			function general_controlClicked(e) {
				var uuid = e.data.uuid;
				switch (structure.controls[uuid].type) {
				case 'Pushbutton':
					sendRequest('jdev/sps/io/'+uuid+'/pulse');
					$(this).addClass('on');
					var el = $(this);
					window.setTimeout(function() {
						el.removeClass('on');
					}, 500);
					break;
				case 'Switch':
					var active = structure.controls[uuid].states.active;
					if (currentValues[active]) {
						sendRequest('jdev/sps/io/'+uuid+'/off');
						//$(this).removeClass('on');
					} else {
						sendRequest('jdev/sps/io/'+uuid+'/on');
						//$(this).addClass('on');
					}
					break;
				}
			}
			
			function general_controlChanged(e) {
				var uuid = e.data.uuid;
				if (valueUuid2Uuid[uuid]) {
					uuid = valueUuid2Uuid[uuid];
				}
				switch(e.currentTarget.tagName) {
				case 'INPUT':
					switch ($(e.currentTarget).attr('type')) {
					case 'text':
						$(e.currentTarget).blur();
						sendRequest('jdev/sps/io/'+uuid+'/'+$(e.currentTarget).val());
						break;
					case 'radio':
						if (e.currentTarget.checked) {
							sendRequest('jdev/sps/io/'+uuid+'/'+$(e.currentTarget).val());
						}
						break;
					case 'range':
						sendRequest('jdev/sps/io/'+uuid+'/'+$(e.currentTarget).val());
						break;
					case 'checkbox':
						sendRequest('jdev/sps/io/'+uuid+'/'+(e.currentTarget.checked?1:0));
						break;
					case 'button':
						break;
					}
					break;
				case 'SELECT':
					var value = $(e.currentTarget).val();
					var control = structure.controls[uuid];
					if (control.type == 'LightController') {
						if (value == 0) {
							value = 'reset';
						} else if (value == 9) {
							value = 'on';
						}
					}
					$(e.currentTarget).blur();
					sendRequest('jdev/sps/io/'+uuid+'/'+value);
					break;
				}
			}
			
			function general_buttonDown(e) {
				var uuid = e.data.uuid;
				if (valueUuid2Uuid[uuid]) {
					uuid = valueUuid2Uuid[uuid];
				}
				$('.'+uuid).each(function() {
					$(this).addClass('active');
				});
				sendRequest('jdev/sps/io/'+uuid+'/1');
			}
			function general_buttonUp(e) {
				var uuid = e.data.uuid;
				if (valueUuid2Uuid[uuid]) {
					uuid = valueUuid2Uuid[uuid];
				}
				$('.'+uuid).each(function() {
					$(this).removeClass('active');
				});
				sendRequest('jdev/sps/io/'+uuid+'/0');
			}
			
			function handle_keypress(e) {
				var uuid = registeredKeys[e.which];
				if (!uuid) {
					return;
				}
				if (valueUuid2Uuid[uuid]) {
					uuid = valueUuid2Uuid[uuid];
				}
				$('.'+uuid).each(function() {
					switch ($(this).attr('type')) {
					case 'text':
						$(this).focus().val(String.fromCharCode(e.which));
						break;
					case 'radio':
						if ($(this).attr('value') == registeredKeyIdx[e.which]+1) {
							$(this).prop('checked', true)
								.trigger('change');
						}
						break;
					case 'range':
						break;
					case 'checkbox':
						this.checked = !this.checked;
						$(this).trigger('change');
						break;
					case 'button':
						if (!$(this).hasClass('active')) {
							$(this).trigger('mousedown');
						}
						break;
					}
				});
				e.stopPropagation();
			}
			
			function handle_keyup(e) {
				var uuid = registeredKeys[e.which];
				if (!uuid) {
					return;
				}
				if (valueUuid2Uuid[uuid]) {
					uuid = valueUuid2Uuid[uuid];
				}
				$('.'+uuid).each(function() {
					switch ($(this).attr('type')) {
					case 'button':
						$(this).trigger('mouseup');
						e.stopPropagation();
						break;
					}
				});
			}
			
			var largechartstats, largechartuuid, largechartvalues, startTs, endTs, dataCount, dataSets;
			var largechartinitialrange, largechartmaxrange, largechartvalueset;
			function showLargeChart(e) {
				$(document.body).prepend($('<div id=largechartoverlay></div>').click(hideLargeChart));
				$(document.body).prepend('<div id=largechart><div id=largechartcanvas class=largechartcanvas></div></div>');
				$('#largechart').prepend($('<div id=closelargechart></div>').click(hideLargeChart)).prepend('<div id=largechartinfo>Drag to pan. Mouse wheel to zoom, or double-click.</div>');
		
				largechartinitialrange = $(e.currentTarget).attr('data-largechartinitialrange');
				largechartinitialrange = largechartinitialrange || 7;
				largechartmaxrange = $(e.currentTarget).attr('data-largechartmaxrange');
				largechartmaxrange = largechartmaxrange || 13;
				largechartvalueset = $(e.currentTarget).attr('data-chartvalueset');
		
				largechartuuid = e.data.uuid;
				
				var valueCount = structure.controls[e.data.uuid].statistic.outputs.length;
				largechartvalues = new Array();
				for (var i=0; i<valueCount; i++) {
					largechartvalues[i] = new Array();
				}
				
				startTs=isNaN, endTs=isNaN;
				dataCount = 0;
				dataSets = 0;
				
				var cmd = 'statistics.json/'+e.data.uuid;
				loadBin(cmd, 'statistics', 'largechartcanvas');
			}
			function gotLargeChartStats() {
				if (largechartstats.Statisticfiles.File.length) {
					for (var i = largechartstats.Statisticfiles.File.length-1; i >= 0; i--) {
						var o = largechartstats.Statisticfiles.File[i];
	
						dataSets++;
						if (dataSets > largechartmaxrange) {
							dataSets--;
							break;
						}
					}
					i++;
					for (; i<largechartstats.Statisticfiles.File.length; i++) {
						var o = largechartstats.Statisticfiles.File[i];
						var cmd = 'binstatisticdata/'+largechartuuid+'/'+o.Year+('0'+o.Month).substr(-2);
						loadBin(cmd, 'largechart', 'largechartcanvas');
					}
				} else {
					var o = largechartstats.Statisticfiles.File;
					var cmd = 'binstatisticdata/'+largechartuuid+'/'+o.Year+('0'+o.Month).substr(-2);
					loadBin(cmd, 'largechart', 'largechartcanvas');
				}
					
				return;
				
			}
			function hideLargeChart(e) {
				$('#largechart').remove();
				$('#largechartoverlay').remove();
			}
			function updateLargeChart() {
				if (++dataCount < dataSets) {
					return;
				}
				
				var now = new Date();
				
				var data = new Array();
				for (var i=0; i<thisBin.valueCount; i++) {
					if (largechartvalueset !== undefined && largechartvalueset != i+1) {
						continue;
					}
					data[largechartvalueset !== undefined ? 0 : i] = {
						color: chartColors[i],
						data: largechartvalues[i]
					}
				}
				var plot = $.plot('.'+thisBin.destination, data, {
					xaxis: {
						show: true,
						zoomRange: [14400000, null],
						panRange: [startTs, endTs],
						mode: 'time',
						min: endTs-largechartinitialrange*24*3600*1000,
						tickFormatter: function(val, axis) {
							var d = new Date(val);
							var axisdays = (axis.max - axis.min) / (24*3600*1000);
							if (axisdays > 10) {
								return strftime('%d/%m/%Y', d);
							} else {
								return strftime('%H:%M<br>%d/%m/%Y', d);
							}
						},
						font: { size: 10, color: '#545454' }
					},
					yaxis: {
						show: true,
						zoomRange: false,
						panRange: false
					},
					series: {
						lines: {
							show: true,
							lineWidth: 1,
							fill: (thisBin.valueCount==1 || largechartvalueset !== undefined),
							fillColor: 'rgba(150, 210, 26, 0.8)', //'#96d21a',
							zero: false
						},
						shadowSize: 0,
						//curvedLines: {
						//	active: true,
						//	apply: true,
						//	monotonicFit: true
						//}
					},
					zoom: {
						interactive: true
					},
					pan: {
						interactive: true
					}
				});
			}
			
			function renderExternalTab(tabid) {
				
				var external = config.external[tabid];
				
				// Clear any unload timeout
				if (unload_externals[tabid]) {
					clearTimeout(unload_externals[tabid]);
					delete unload_externals[tabid];
				}
				
				// Always send WOL even if already rendered
				if (external.wol) {
					if (!external.wol.uuid) {
						var uuid = findControlUuid(external.wol.control, external.room, external.cat);
						if (!uuid) {
							return;
						}
						config.external[tabid].wol.uuid = uuid;
					}
					sendRequest('jdev/sps/io/'+external.wol.uuid+'/pulse');
				}
				
				if (!rendered_externals[tabid]) {
					var html = '<div id='+tabid+' class=tab><iframe class=external src="'+external.url+'"></iframe></div>';
					$('#'+tabid).remove();
					$('#topbar').after(html);
					
					rendered_externals[tabid] = true;
					$('#'+tabid).show();
				}
			}
					
			
			function findControlUuid(controlName, roomName, catName, suppresserrors) {
				if (roomName) {
					var roomuuid = roomName2Uuid[roomName];
				} else {
					var roomuuid = '';
				}
				if (catName) {
					var catuuid = catName2Uuid[catName];
				} else {
					var catuuid = '';
				}
				for (var uuid in structure.controls) {
					if (structure.controls[uuid].name == controlName
					   && (roomuuid == '' || structure.controls[uuid].room == roomuuid)
					   && (catuuid == '' || structure.controls[uuid].cat == catuuid)) {
					   	return uuid;
					}
				}
				if (typeof suppresserrors == 'undefined') {
					errorInfo("Unknown control '"+controlName+"' / room '"+roomName+"' / cat '"+catName+"'");
				}
				return '';
			}
			
			
			/*
			Queue of binary objects to retrieve
			Each element is an object:
			{
				cmd: <command to send>,
				type: 'minichart|largechart|svg',
				destination: <class of html element(s) to insert chart or svg>,
				
				uuid: <uuid of control>,			// for charts only
				valueCount: <number of statistical value sets>	// for charts only
			}
			*/
			var binQueue = new Array();
			var binQueueActive = false;
			var thisBin = null;
			function loadBin(cmd, type, destination) {
				if (matches = cmd.match(/^binstatisticdata\/([0-9a-f-]+)\/([0-9][0-9][0-9][0-9])([0-9][0-9])([0-9][0-9])$/)) {
					var uuid = matches[1];
					var date = [matches[2],matches[3],matches[4]];
					var valueCount = structure.controls[uuid].statistic.outputs.length;
				} else if (matches = cmd.match(/^binstatisticdata\/([0-9a-f-]+)\/([0-9][0-9][0-9][0-9])([0-9][0-9])$/)) {
					var uuid = matches[1];
					var date = [matches[2],matches[3]];
					var valueCount = structure.controls[uuid].statistic.outputs.length;
				} else {
					var uuid = '';
					var date = '';
					var valueCount = 0;
				}
				binQueue.push({cmd: cmd, type: type, destination: destination, uuid: uuid, date: date, valueCount: valueCount});
				if (!binQueueActive) {
					binQueueActive = true;
					nextBin();
				}
			}
			var binTimer = null;
			function nextBin() {
				clearTimeout(binTimer);
				thisBin = binQueue.shift();
				if (typeof thisBin == 'undefined') {
					binQueueActive = false;
					thisBin = null;
				} else {
					binTimer = setTimeout(function() {
						//alert('Timeout with bin request: '+thisBin.cmd);
						nextBin();
					}, 5000);
					sendRequest(thisBin.cmd, gotBin);
				}
			}
			function updateCharts(bClearExistingCharts) {
				if (bClearExistingCharts) {
					binQueue = new Array();
					for (var uuid in chartsToKeepUpdated) {
						$('.'+chartsToKeepUpdated[uuid]).empty();
					}
				}
				var now = new Date();
				if (chartDateOffset[activeTab] != 0) {
					now.setDate(now.getDate()+chartDateOffset[activeTab]);
				}
				var day = ('0'+now.getDate()).substr(-2);
				var month = ('0'+(now.getMonth()+1)).substr(-2);
				var year = now.getFullYear().toString();
				var chartDate = year+month+day;
				for (var uuid in chartsToKeepUpdated) {
					var cmd = 'binstatisticdata/'+uuid+'/'+chartDate;
					loadBin(cmd, 'minichart', chartsToKeepUpdated[uuid]);
				}
			}
			
			function noHandler(msg) {
				//log('noHandler');
				//log(msg);
			}
			function gotBin(binData) {
				//log('gotBin');
			
				if (typeof binData == 'string') {
					switch (thisBin.type) {
					case 'svg':
						$('.'+thisBin.destination).html(binData);
						break;
					case 'statistics':
						largechartstats = JSON.parse(binData);
						gotLargeChartStats();
						break;
					}
					nextBin();
				} else {
					var matches;
					switch (thisBin.type) {
					case 'minichart':
						var digiChartResolution = 12;
						var ptr = 0;
						var chartvalues = new Array();
						var startHour = 0;
						for (var i=0; i<thisBin.valueCount; i++) {
							chartvalues[i] = new Array();
						}
						if (binData) {
							try {
								var dataview = new DataView(binData);
							} catch(error) {
								log(error);
							}
							while (ptr < dataview.byteLength) {
								// ts
								var ts = dataview.getUint32(ptr, true);
								var dayHourMin = tsToDayHourMin(ts);
								
								// Exit if wrong date (either looped to next day, or Loxone bug where is sends wrong data)
								if (dayHourMin[0] != parseInt(thisBin.date[2])) {
									break;
								}
								
								var hour = dayHourMin[1];
								ptr+=4;
								
								// dVal
								for (var i=0; i<thisBin.valueCount; i++) {
									var dVal = dataview.getFloat64(ptr, true);
									ptr+=8;
									
									// Digital stats
									if (structure.controls[thisBin.uuid].type == 'InfoOnlyDigital') {
										var min = dayHourMin[2];
										// Timestamp to quantised index
										var qidx = Math.floor((hour+min/60)*digiChartResolution);
										if ((chartvalues[i][qidx] === undefined || chartvalues[i][qidx] == 0) && dVal == 1) {
											chartvalues[i][qidx] = 1;
										}
									// Analog stats
									} else {
										chartvalues[i][hour] = dVal.toFixed(2);
									}
								}
								
								//log(ptr+'/'+dataview.byteLength + ' | ' + tsToDateTime(ts) + ' = ' +dVal);
							};
						}
						
						var currentTime = new Date(
							currentValues[structure.globalStates.miniserverTime].substr(0,4),
							parseInt(currentValues[structure.globalStates.miniserverTime].substr(5,2)),
							parseInt(currentValues[structure.globalStates.miniserverTime].substr(8,2)),
							parseInt(currentValues[structure.globalStates.miniserverTime].substr(11,2)),
							parseInt(currentValues[structure.globalStates.miniserverTime].substr(14,2)),
							parseInt(currentValues[structure.globalStates.miniserverTime].substr(17,2))
						);
						var secsSinceLoad = ((new Date()).getTime() - timeOfLoad)/1000;
						currentTime.setSeconds(currentTime.getSeconds()+secsSinceLoad);
						
						if (ts) {
							var endHour = hour;
							if (endHour <= 0) {
								endHour += 24;
							}
							var endMin = 59;
							if (chartDateOffset[activeTab] == 0) {
								var currentMin = currentTime.getMinutes();
								if (endMin > currentMin) {
									endMin = currentMin;
								}
							}
						}
						
						// For digital stats, always show full period until now, or full day if past day
						if (structure.controls[thisBin.uuid].type == 'InfoOnlyDigital') {
							if (chartDateOffset[activeTab] == 0) {
								var hour = currentTime.getHours();
								var min = currentTime.getMinutes();
							} else {
								var hour = 23;
								var min = 59;
							}
							// Timestamp to quantised index
							var qidx = Math.floor((hour+min/60)*digiChartResolution);
							// Override startHour & endHour
							startHour = 0;
							endHour = hour;
							endMin = min;
							// Set zero value for any unpopulated hours (otherwise sparkline doesn't like it)
							for (var i=0; i<thisBin.valueCount; i++) {
								for (var j=0; j<=qidx; j++) {
									if (chartvalues[i][j] === undefined) {
										chartvalues[i][j] = 0;
									}
								}
							}
						} else {
							// Set zero value for any unpopulated hours (otherwise sparkline doesn't like it)
							for (var i=0; i<thisBin.valueCount; i++) {
								for (var j=0; j<=endHour; j++) {
									if (chartvalues[i][j] === undefined) {
										chartvalues[i][j] = 0;
									}
								}
							}
						}
					
						// Append current value (for analog stats only on current day)
						if (thisBin.valueCount == 1 && structure.controls[thisBin.uuid].type != 'InfoOnlyDigital' && chartDateOffset[activeTab] == 0) {
							if (currentValues.hasOwnProperty(thisBin.uuid)) {
								chartvalues[0][chartvalues[0].length] = currentValues[thisBin.uuid].toFixed(2);
							}
						}
						
						var chartvalueset = $('.'+thisBin.destination).attr('data-chartvalueset');
						var chartwidth = $('.'+thisBin.destination).attr('data-chartwidth');
						if (chartwidth === undefined) {
							chartwidth = '75';
						}
						for (var i=0; i<thisBin.valueCount; i++) {
							if (chartvalueset !== undefined && chartvalueset != i+1) {
								continue;
							}
							// Get symbol suffix
							var sym = '';
							if (structure.controls[thisBin.uuid].type == 'InfoOnlyDigital') {
								var tooltipFormat = null;
								var tooltipFormatter = function (sparkline, options, fields) {
									var hour = Math.floor(fields.x/digiChartResolution);
									var mins = Math.round(fields.x%digiChartResolution*60/digiChartResolution);
									return hour+':'+(mins<10?'0':'')+mins;
								}
							} else if (matches = structure.controls[thisBin.uuid].statistic.outputs[i].format.match(/%.([0-9]+)f(.*)/)) {
								var tooltipFormat = '{{x}}:00 &rarr; {{y}}'+matches[2].replace('%%', '%');
								var tooltipFormatter = null;
							} else {
								var tooltipFormat = '{{x}}:00 &rarr; {{y}}';
								var tooltipFormatter = null;
							}
							$('.'+thisBin.destination).sparkline(chartvalues[i], {
								type: 'line',
								width: chartwidth+'px',
								lineColor: chartColors[i],
								fillColor: (thisBin.valueCount==1?'#96d21a':''),
								spotRadius: 0,
								tooltipFormat: tooltipFormat,
								tooltipFormatter: tooltipFormatter,
								composite: (i>0 && chartvalueset === undefined)
							});
						}
						if (endHour) {
							$('#'+activeTab+' .chart-time-range').html(startHour+':00-'+endHour+':'+(endMin<10?'0':'')+endMin);
							var prelink = $('<a href="javascript:void(0);">&lArr;</a>').click(
								function() {
									chartDateOffset[activeTab]--;
									updateCharts(true);
								} );
							var postlink = $('<a href="javascript:void(0);">&rArr;</a>').click(
								function() {
									if (chartDateOffset[activeTab]<0) {
										chartDateOffset[activeTab]++;
										updateCharts(true);
									}
								} );
							$('#'+activeTab+' .chart-controls').html(prelink).append('&nbsp; '+thisBin.date[2]+'/'+thisBin.date[1]+'/'+thisBin.date[0]+'&nbsp; '+startHour+':00-'+endHour+':'+(endMin<10?'0':'')+endMin+'&nbsp; ').append(postlink);
						}
						break;
						
					case 'largechart':
						var ptr = 0;
						var dataview = new DataView(binData);
						while (ptr < dataview.byteLength) {
							// ts
							var ts = dataview.getUint32(ptr, true);
							ts = loxts2jsts(ts);
							if (isNaN(startTs) || ts < startTs) {
								startTs = ts;
							}
							if (isNaN(endTs) || ts > endTs) {
								endTs = ts;
							}
							ptr+=4;
							
							// dVal
							for (var i=0; i<thisBin.valueCount; i++) {
								var dVal = dataview.getFloat64(ptr, true);
								ptr+=8;
								largechartvalues[i][largechartvalues[i].length] = [ts, dVal, -1000];
							}
						};
						updateLargeChart();
						break;
					}
					nextBin();
				}
			}
					
			
			function loxts2jsts(ts) {
				return (ts+1230768000-timezoneOffset*3600)*1000;
			}
				
			
			function uuidBinToString(dataview, offset) {
				var uuid = '';
				uuid += ('00000000'+dataview.getUint32(offset, true).toString(16)).substr(-8);
				uuid += '-';
				uuid += ('0000'+dataview.getUint16(offset+4, true).toString(16)).substr(-4);
				uuid += '-';
				uuid += ('0000'+dataview.getUint16(offset+6, true).toString(16)).substr(-4);
				uuid += '-';
				uuid += ('00'+dataview.getUint8(offset+8, true).toString(16)).substr(-2);
				uuid += ('00'+dataview.getUint8(offset+9, true).toString(16)).substr(-2);
				uuid += ('00'+dataview.getUint8(offset+10, true).toString(16)).substr(-2);
				uuid += ('00'+dataview.getUint8(offset+11, true).toString(16)).substr(-2);
				uuid += ('00'+dataview.getUint8(offset+12, true).toString(16)).substr(-2);
				uuid += ('00'+dataview.getUint8(offset+13, true).toString(16)).substr(-2);
				uuid += ('00'+dataview.getUint8(offset+14, true).toString(16)).substr(-2);
				uuid += ('00'+dataview.getUint8(offset+15, true).toString(16)).substr(-2);
				return uuid;
			};
			
			function decodeUTF8(dataview, offset, length) {
				var s = '';
				var i = 0;
				while (i < length) {
					var c = dataview.getUint8(offset + i++, true);
					switch(c >> 4) {
					case 0: case 1: case 2: case 3: case 4: case 5: case 6: case 7:
						s += String.fromCharCode(c);
						break;
					case 12: case 13:
						c2 = dataview.getUint8(offset + i++, true);
						s += String.fromCharCode(((c & 0x1F) << 6) | (c2 & 0x3F));
						break;
					case 14:
						c2 = dataview.getUint8(offset + i++, true);
						c3 = dataview.getUint8(offset + i++, true);
						s += String.fromCharCode(((c & 0x0F) << 12) | ((c2 & 0x3F) << 6) | ((c3 & 0x3F) << 0));
						break;
					}
				}
				return s;
			}
			
			function tsToDayHourMin(ts) {
				var date = new Date((ts+1230768000)*1000);
				date.setHours(date.getHours()-timezoneOffset);
				
				return [date.getDate(), date.getHours(), date.getMinutes()];
			}
			function tsToDateTime(ts) {
				var date = new Date((ts+1230768000)*1000);
				date.setHours(date.getHours()-timezoneOffset);
				
				var day = date.getDate();
				var month = date.getMonth()+1;
				var year = date.getFullYear();
				
				var hours = date.getHours();
				var minutes = "0" + date.getMinutes();
				var seconds = "0" + date.getSeconds();
				
				return day+'/'+month+'/'+year+' '+ hours + ':' + minutes.substr(-2) + ':' + seconds.substr(-2);
			}
			
			function sendRequest(msg, callback) {
				callback = callback || noHandler;
				try{
					socket.send(msg).then(callback);
					log('Sent: '+msg)
				
				} catch(exception){
					log('Error: '+exception);
				}
			}
			
			function errorInfo(msg) {
				$('#errorsmsg').show();
				$('#errors').append('<p>'+msg+'</p>');
			}
			function debugOut(msg) {
				$('#debugout').show();
				$('#debugout').prepend('<p>'+msg+'</p>');
			}
			
			function log(msg) {
				console.log(msg);
			}
			
			var controlElements = {
				'switch':	$('<svg width="30" height="30" viewBox="0 0 30 30" class="lightswitch control"><path d="M15.013,1c0.413,0,0.749,0.338,0.749,0.753v3.013c0,0.415-0.336,0.753-0.749,0.753c-0.413,0-0.749-0.338-0.749-0.753V1.753 C14.264,1.338,14.6,1,15.013,1 M5.462,4.97c0.292-0.294,0.767-0.294,1.059,0l2.117,2.13c0.292,0.294,0.292,0.772,0,1.065 c-0.292,0.294-0.767,0.294-1.059,0L5.462,6.036C5.17,5.742,5.17,5.264,5.462,4.97 M1.499,14.573c0-0.415,0.336-0.753,0.748-0.753 h2.994c0.413,0,0.749,0.338,0.749,0.753c0,0.415-0.336,0.753-0.749,0.753H2.248C1.835,15.326,1.499,14.988,1.499,14.573 M5.445,24.183c-0.292-0.294-0.292-0.771,0-1.065l2.117-2.13c0.292-0.294,0.767-0.294,1.059,0c0.292,0.294,0.292,0.772,0,1.065 l-2.117,2.131C6.211,24.477,5.736,24.477,5.445,24.183 M24.538,24.201c-0.292,0.294-0.767,0.294-1.059,0l-2.117-2.131 c-0.292-0.294-0.292-0.772,0-1.065c0.292-0.294,0.767-0.294,1.059,0l2.117,2.131C24.83,23.43,24.83,23.907,24.538,24.201 M28.501,14.598c0,0.415-0.336,0.753-0.748,0.753h-2.994c-0.413,0-0.748-0.338-0.748-0.753c0-0.415,0.336-0.753,0.748-0.753h2.994 C28.165,13.845,28.501,14.183,28.501,14.598 M24.555,4.988c0.292,0.294,0.292,0.772,0,1.065l-2.117,2.13 c-0.292,0.294-0.767,0.294-1.059,0c-0.292-0.294-0.292-0.771,0-1.065l2.117-2.13C23.789,4.694,24.264,4.694,24.555,4.988 M15.001,28.996c1.652,0,2.993-1.349,2.993-3.013h-5.988C12.006,27.646,13.346,28.996,15.001,28.996z M14.806,17.276 c0.131,0.06,0.283,0.06,0.413,0l1.639-0.749l-0.922,3.168c0.402-0.076,0.784-0.203,1.147-0.37l1.038-3.568 c0.056-0.188-0.004-0.392-0.151-0.52c-0.148-0.129-0.357-0.159-0.534-0.078l-2.424,1.108l-2.424-1.108 c-0.179-0.081-0.388-0.05-0.537,0.081c-0.147,0.131-0.205,0.337-0.146,0.526l1.117,3.604c0.362,0.159,0.749,0.265,1.149,0.334 l-0.982-3.167L14.806,17.276z M15.012,7.214c-4.134,0-7.485,3.372-7.485,7.533c0,2.228,0.955,4.224,2.483,5.603v3.627 c0,0.555,0.509,1.004,1.121,1.004h7.763c0.612,0,1.097-0.45,1.097-1.004V20.35c1.528-1.379,2.508-3.375,2.508-5.603 C22.498,10.587,19.147,7.214,15.012,7.214z M15.012,20.633c-3.302,0-5.988-2.703-5.988-6.026s2.686-6.026,5.988-6.026 s5.988,2.703,5.988,6.026S18.314,20.633,15.012,20.633z"/></svg>'),
				'dimmer':	$('<svg width="30" height="124" viewBox="-9 -2 30 100" class="fader control">\
							<rect class="faderBackground" width="12" height="100" rx="2"/>\
							<rect class="faderForeground" width="12" height="50" rx="2"/>\
							<circle class="faderKnob" cy="50" cx="6" r="10"/>\
						</svg>'),
				'button':	$('<svg viewBox="0 0 50 50" class="button">\
							<circle class="bg" cx="25" cy="25" r="25"/>\
							<path class="ring" d="M25,15.5c5.2,0,9.5,4.3,9.5,9.5s-4.3,9.5-9.5,9.5s-9.5-4.3-9.5-9.5S19.8,15.5,25,15.5 M25,12 c-7.2,0-13,5.8-13,13s5.8,13,13,13s13-5.8,13-13S32.2,12,25,12L25,12z"/>\
						</svg>')
			};
			
			var chartColors = [
				'#679012',
				'#144590',
				'#901b8c',
				'#904919',
				'#90132d',
				'#d1cf1c',
				'#2ecfd1',
				'#26d12a'
			];
		}
	}
});